# -*- python -*-
# $RCSfile: pixelgroupwidget.py,v $
# $Revision: 1.29 $
# $Author: rlua $
# $Date: 2007-09-28 03:20:39 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from ooflib.SWIG.common import switchboard
from ooflib.common import debug
from ooflib.common.IO import selectionplaceholder
from ooflib.common.IO import pixelgroupparam
from ooflib.common.IO.GUI import chooser
from ooflib.common.IO.GUI import parameterwidgets
from ooflib.common.IO.GUI import whowidget
import ooflib.common.microstructure

# Widget for choosing a pixel group.  It *requires* a non trivial
# scope argument (WidgetScope object) so that it can find a WhoWidget
# for a microstructure.

class PixelGroupWidget(parameterwidgets.ParameterWidget):
    def __init__(self, param, scope=None, name=None):
        self.groupchooser = chooser.ChooserWidget([], name=name)
        parameterwidgets.ParameterWidget.__init__(self, self.groupchooser.gtk,
                                                  scope=scope)
        self.mswidget = self.scope.findWidget(
            lambda w: isinstance(w, whowidget.WhoWidget)
            and w.whoclass is ooflib.common.microstructure.microStructures)
        self.update()
        if param.value is not None:
            self.set_value(param.value)
        self.sbcallbacks = [
            switchboard.requestCallbackMain(self.mswidget, self.update),
            switchboard.requestCallbackMain("new pixel group", self.update),
            switchboard.requestCallbackMain("renamed pixel group",
                                            self.update),
            switchboard.requestCallbackMain("destroy pixel group",
                                            self.update),
            switchboard.requestCallbackMain(self.groupchooser, self.chooserCB)
            ]
    def cleanUp(self):
        map(switchboard.removeCallback, self.sbcallbacks)
        self.mswidget = None
        parameterwidgets.ParameterWidget.cleanUp(self)
    def update(self, *args, **kwargs):
        msname = self.mswidget.get_value()
        if msname:
            ms = ooflib.common.microstructure.microStructures[msname]
            if ms:
                names = ms.getObject().groupNames()
                self.groupchooser.update(names)
                self.widgetChanged(len(names) > 0, interactive=0)
                return
        self.groupchooser.update([])
        self.widgetChanged(0, interactive=0)
    def chooserCB(self):
        self.widgetChanged(self.groupchooser.nChoices() > 0, interactive=1)
    def get_value(self):                # returns *name* of group
        return self.groupchooser.get_value()
    def set_value(self, groupname):
        self.groupchooser.set_state(groupname)
        

def _makePixelGroupWidget(self, scope=None):
    return PixelGroupWidget(self, scope=scope, name=self.name)

pixelgroupparam.PixelGroupParameter.makeWidget = _makePixelGroupWidget

############

class PixelAggregateWidget(PixelGroupWidget):
    def update(self, *args, **kwargs):
        msname = self.mswidget.get_value()
        if msname:
            ms = ooflib.common.microstructure.microStructures[msname]
            if ms:
                self.groupchooser.update(
                    [selectionplaceholder.selectionIDstring,
                     selectionplaceholder.allIDstring] +
                    ms.getObject().groupNames())
                self.widgetChanged(1, interactive=0)
                return
        self.groupchooser.update([])
        self.widgetChanged(0, interactive=0)
    def get_value(self):
        rval = self.groupchooser.get_value()
        if rval == selectionplaceholder.selectionIDstring:
            return selectionplaceholder.selectionPlaceHolder
        if rval == selectionplaceholder.allIDstring:
            return selectionplaceholder.allPlaceHolder
        return rval

def _makePixelAggregateWidget(self, scope=None):
    return PixelAggregateWidget(self, scope=scope, name=self.name)

pixelgroupparam.PixelAggregateParameter.makeWidget = _makePixelAggregateWidget

#Interface branch
class PixelGroupInterfaceWidget(PixelGroupWidget):
    def update(self, *args, **kwargs):
        msname = self.mswidget.get_value()
        if msname:
            ms = ooflib.common.microstructure.microStructures[msname]
            if ms:
                names = ms.getObject().groupNames()
                self.groupchooser.update(names+["<No pixelgroup>"])
                self.widgetChanged(len(names) > 0, interactive=0)
                return
        self.groupchooser.update([])
        self.widgetChanged(0, interactive=0)        

def _makePixelGroupInterfaceWidget(self, scope=None):
    return PixelGroupInterfaceWidget(self, scope=scope, name=self.name)

pixelgroupparam.PixelGroupInterfaceParameter.makeWidget = _makePixelGroupInterfaceWidget
