# -*- python -*-
# $RCSfile: materialwidget.py,v $
# $Revision: 1.17 $
# $Author: rlua $
# $Date: 2007-10-30 22:21:30 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from ooflib.SWIG.common import switchboard
from ooflib.SWIG.engine import material
from ooflib.common import debug
from ooflib.common.IO.GUI import chooser
from ooflib.common.IO.GUI import parameterwidgets
from ooflib.common.IO.GUI import whowidget
from ooflib.engine import materialmanager
from ooflib.engine import mesh
from ooflib.engine import subproblemcontext
from ooflib.engine.IO import materialmenu
from ooflib.engine.IO import materialparameter
#Interface branch
import ooflib.common.microstructure

class MaterialWidget(parameterwidgets.ParameterWidget):
    def __init__(self, param, scope=None, name=None):
        self.chooser = chooser.ChooserWidget([], callback=self.chooserCB,
                                             name=name)
        parameterwidgets.ParameterWidget.__init__(self, self.chooser.gtk, scope)
        self.sbcallbacks = [
            switchboard.requestCallbackMain("new_material", self.update),
            switchboard.requestCallbackMain("remove_material", self.update)
            ]
        self.update()
        if param.value is not None:
            self.set_value(param.value)
    def chooserCB(self, gtkobj, name):
        self.widgetChanged(validity=self.chooser.nChoices()>0, interactive=1)
    def cleanUp(self):
        map(switchboard.removeCallback, self.sbcallbacks)
        parameterwidgets.ParameterWidget.cleanUp(self)
    def update(self, *args):
        names = materialmanager.getMaterialNames()
        names.sort()
        self.chooser.update(names)
        self.widgetChanged(len(names) > 0, interactive=0)
    def get_value(self):
        return self.chooser.get_value()
    def set_value(self, material):
        self.chooser.set_state(material)
        self.widgetChanged(validity=(material is not None), interactive=0)

def _MaterialParameter_makeWidget(self, scope=None):
    return MaterialWidget(self, scope, name=self.name)
materialparameter.MaterialParameter.makeWidget = _MaterialParameter_makeWidget

#Interface branch
class InterfaceMaterialWidget(MaterialWidget):
    def update(self, *args):
        names = materialmanager.getInterfaceMaterialNames()
        names.sort()
        self.chooser.update(names)
        self.widgetChanged(len(names) > 0, interactive=0)

def _InterfaceMaterialParameter_makeWidget(self, scope=None):
    return InterfaceMaterialWidget(self, scope, name=self.name)
materialparameter.InterfaceMaterialParameter.makeWidget = _InterfaceMaterialParameter_makeWidget

#Interface branch
class BulkMaterialWidgetExtra(MaterialWidget):
    def update(self, *args):
        names = materialmanager.getBulkMaterialNames()
        names.sort()
        self.chooser.update(names+["<No material>"])
        self.widgetChanged(len(names) > 0, interactive=0)

def _BulkMaterialParameterExtra_makeWidget(self, scope=None):
    return BulkMaterialWidgetExtra(self, scope, name=self.name)
materialparameter.BulkMaterialParameterExtra.makeWidget = _BulkMaterialParameterExtra_makeWidget

#Interface branch
class BulkMaterialWidget(MaterialWidget):
    def update(self, *args):
        names = materialmanager.getBulkMaterialNames()
        names.sort()
        self.chooser.update(names)
        self.widgetChanged(len(names) > 0, interactive=0)

def _BulkMaterialParameter_makeWidget(self, scope=None):
    return BulkMaterialWidget(self, scope, name=self.name)
materialparameter.BulkMaterialParameter.makeWidget = _BulkMaterialParameter_makeWidget

class MeshMaterialWidget(MaterialWidget):
    def __init__(self, param, scope=None, name=None):
        self.meshwidget = scope.findWidget(
            lambda w: isinstance(w, whowidget.WhoWidget)
            and w.whoclass in (mesh.meshes, subproblemcontext.subproblems))
        MaterialWidget.__init__(self, param, scope, name)
        self.sbcallbacks.append(
            switchboard.requestCallbackMain(self.meshwidget, self.update))
    def update(self, *args):
        meshname = self.meshwidget.get_value(depth=3)
        meshctxt = mesh.meshes[meshname]
        matls = material.getMaterials(meshctxt.getMicrostructure()) + \
                meshctxt.explicitMaterials()
        names = [m.name() for m in matls]
        names.sort()
        self.chooser.update(names)
        self.widgetChanged(len(names) > 0, interactive=0)

def _MeshMatParam_makeWidget(self, scope=None):
    return MeshMaterialWidget(self, scope, name=self.name)
materialparameter.MeshMaterialParameter.makeWidget = _MeshMatParam_makeWidget
        
########################

class AnyMaterialWidget(MaterialWidget):
    def update(self, *args):
        names = materialmanager.getMaterialNames()
        names.sort()
        self.chooser.update(materialparameter.AnyMaterialParameter.extranames
                            + names)
        self.widgetChanged(validity=1, interactive=0)

def _AnyMaterialParameter_makeWidget(self, scope=None):
    return AnyMaterialWidget(self, scope, name=self.name)

materialparameter.AnyMaterialParameter.makeWidget = _AnyMaterialParameter_makeWidget

#Interface branch
class InterfaceAnyMaterialWidget(MaterialWidget):
    def update(self, *args):
        names = materialmanager.getInterfaceMaterialNames()
        names.sort()
        self.chooser.update(materialparameter.InterfaceAnyMaterialParameter.extranames
                            + names)
        self.widgetChanged(validity=1, interactive=0)

def _InterfaceAnyMaterialParameter_makeWidget(self, scope=None):
    return InterfaceAnyMaterialWidget(self, scope, name=self.name)

materialparameter.InterfaceAnyMaterialParameter.makeWidget = _InterfaceAnyMaterialParameter_makeWidget

########################
        
class MaterialsWidget(parameterwidgets.ParameterWidget):
    def __init__(self, param, scope=None, name=None):
        names = materialmanager.getMaterialNames()
        names.sort()
        self.widget = chooser.ScrolledMultiListWidget(names,
                                                      callback=self.widgetCB)
        parameterwidgets.ParameterWidget.__init__(self, self.widget.gtk, scope,
                                                  name=name)
        self.widget.set_selection(param.value)
        self.sbcallbacks = [
            switchboard.requestCallbackMain('new_material', self.newMaterial),
            switchboard.requestCallbackMain('remove_material',
                                            self.newMaterial)
            ]
        self.widgetChanged((param.value is not None), interactive=0) 
    def cleanUp(self):
        map(switchboard.removeCallback, self.sbcallbacks)
        parameterwidgets.ParameterWidget.cleanUp(self)
    def newMaterial(self, *args):
        names = materialmanager.getMaterialNames()
        names.sort()
        self.widget.update(names)
    def get_value(self):
        return self.widget.get_value()
    def set_value(self, value):
        self.widget.set_selection(value)
    def widgetCB(self, list, interactive):
        self.widgetChanged(len(list) > 0, interactive=1)

def _MaterialsWidget_makeWidget(self, scope=None):
    return MaterialsWidget(self, scope, name=self.name)

materialparameter.ListOfMaterialsParameter.makeWidget = _MaterialsWidget_makeWidget

##################################################################
#Interface branch
#The following widgets list interface names

#Listbox
class ListOfInterfacesWidget(parameterwidgets.ParameterWidget):
    def __init__(self, param, scope=None, name=None):
        self.widget = chooser.ScrolledMultiListWidget([],
                                                      callback=self.widgetCB)

        parameterwidgets.ParameterWidget.__init__(self, self.widget.gtk,
                                                  scope=scope, name=name)
        self.mswidget = self.scope.findWidget(
            lambda w: isinstance(w, whowidget.WhoWidget)
            and w.whoclass is ooflib.common.microstructure.microStructures)
        self.update()
        if param.value is not None:
            self.widget.set_selection(param.value)
        self.sbcallbacks = [
            switchboard.requestCallbackMain(self.mswidget, self.update)
            ]
    def cleanUp(self):
        map(switchboard.removeCallback, self.sbcallbacks)
        self.mswidget = None
        #self.interfacewidget = None
        parameterwidgets.ParameterWidget.cleanUp(self)
    def update(self, *args, **kwargs):
        msname = self.mswidget.get_value()
        if msname:
            ms = ooflib.common.microstructure.microStructures[msname]
            if ms:
                interfacemsplugin=ms.getObject().getPlugIn("Interfaces")
                names=interfacemsplugin.getInterfaceNames()
                names.sort()
                self.widget.update(names)
                self.widgetChanged(len(names) > 0, interactive=0)
                return
        self.widget.update([])
        self.widgetChanged(0, interactive=0)
    def get_value(self):
        return self.widget.get_value()
    def set_value(self, value):
        self.widget.set_selection(value)
    def widgetCB(self, list, interactive):
        self.widgetChanged(len(list) > 0, interactive=1)

def _makeListOfInterfacesWidget(self, scope=None):
    return ListOfInterfacesWidget(self, scope=scope, name=self.name)

materialparameter.ListOfInterfacesParameter.makeWidget = _makeListOfInterfacesWidget

#Listbox. This widget differs from the above widget by having
#a self.interfacewidget member
class ListOfInterfacesWithMaterialWidget(parameterwidgets.ParameterWidget):
    def __init__(self, param, scope=None, name=None):
        self.widget = chooser.ScrolledMultiListWidget([],
                                                      callback=self.widgetCB)

        parameterwidgets.ParameterWidget.__init__(self, self.widget.gtk,
                                                  scope=scope, name=name)
        self.mswidget = self.scope.findWidget(
            lambda w: isinstance(w, whowidget.WhoWidget)
            and w.whoclass is ooflib.common.microstructure.microStructures)
        self.interfacewidget = self.scope.findWidget(
            lambda w: isinstance(w, InterfaceMaterialWidget))
        self.update()
        if param.value is not None:
            self.widget.set_selection(param.value)
        self.sbcallbacks = [
            switchboard.requestCallbackMain(self.mswidget, self.update),
            switchboard.requestCallbackMain(self.interfacewidget, self.update)
            ]
    def cleanUp(self):
        map(switchboard.removeCallback, self.sbcallbacks)
        self.mswidget = None
        self.interfacewidget = None
        parameterwidgets.ParameterWidget.cleanUp(self)
    def update(self, *args, **kwargs):
        msname = self.mswidget.get_value()
        interfacematname = self.interfacewidget.get_value()
        if msname:
            ms = ooflib.common.microstructure.microStructures[msname]
            if ms:
                interfacemsplugin=ms.getObject().getPlugIn("Interfaces")
                names=interfacemsplugin.getInterfaceNamesWithMaterial(interfacematname)
                if names:
                    names.sort()
                    self.widget.update(names)
                    self.widgetChanged(len(names) > 0, interactive=0)
                    return
        self.widget.update([])
        self.widgetChanged(0, interactive=0)
    def get_value(self):
        return self.widget.get_value()
    def set_value(self, value):
        self.widget.set_selection(value)
    def widgetCB(self, list, interactive):
        self.widgetChanged(len(list) > 0, interactive=1)

def _makeListOfInterfacesWithMaterialWidget(self, scope=None):
    return ListOfInterfacesWithMaterialWidget(self, scope=scope, name=self.name)

materialparameter.ListOfInterfacesWithMaterialParameter.makeWidget = _makeListOfInterfacesWithMaterialWidget

#Drop-down combobox
class InterfacesWidget(parameterwidgets.ParameterWidget):
    def __init__(self, param, scope=None, name=None):
        self.widget = chooser.ChooserWidget([],
                                            callback=self.widgetCB,
                                            name=name)
        parameterwidgets.ParameterWidget.__init__(self, self.widget.gtk,
                                                  scope=scope, name=name)
        self.mswidget = self.scope.findWidget(
            lambda w: isinstance(w, whowidget.WhoWidget)
            and w.whoclass is mesh.meshes) #meshwidget
        self.sbcallbacks = [
            switchboard.requestCallbackMain(self.mswidget, self.update)
            ]
        self.update(interactive=0)
        if param.value is not None and self.widget.nChoices()>0:
            self.set_value(param.value)
    def cleanUp(self):
        map(switchboard.removeCallback, self.sbcallbacks)
        self.mswidget = None
        #self.interfacewidget = None
        parameterwidgets.ParameterWidget.cleanUp(self)
    def update(self, *args, **kwargs):
        msname = self.mswidget.get_value(depth=1) #Get the ms path
        if msname:
            ms = ooflib.common.microstructure.microStructures[msname]
            if ms:
                interfacemsplugin=ms.getObject().getPlugIn("Interfaces")
                names=interfacemsplugin.getInterfaceNames()
                names.sort()
                self.widget.update(names)
                self.widgetChanged(len(names) > 0, interactive=0)
                return
        self.widget.update([])
        self.widgetChanged(0, interactive=0)
    def get_value(self):
        return self.widget.get_value()
    def set_value(self, value):
        self.widget.set_state(value)
    def widgetCB(self, gtkobj, name):
        self.widgetChanged(validity=self.widget.nChoices()>0, interactive=1)

def _makeInterfacesWidget(self, scope=None):
    return InterfacesWidget(self, scope=scope, name=self.name)

materialparameter.InterfacesParameter.makeWidget = _makeInterfacesWidget

## Includes skeleton boundary names ##
#Listbox
class ListOfInterfacesSkelBdyWidget(ListOfInterfacesWidget):
    def update(self, *args, **kwargs):
        msname = self.mswidget.get_value()
        if msname:
            ms = ooflib.common.microstructure.microStructures[msname]
            if ms:
                interfacemsplugin=ms.getObject().getPlugIn("Interfaces")
                inames=interfacemsplugin.getInterfaceNames()
                sbnames=interfacemsplugin.getSkelBdyNames()
                inames.sort()
                sbnames.sort()
                names=inames+sbnames
                self.widget.update(names)
                self.widgetChanged(len(names) > 0, interactive=0)
                return
        self.widget.update([])
        self.widgetChanged(0, interactive=0)

def _makeListOfInterfacesSkelBdyWidget(self, scope=None):
    return ListOfInterfacesSkelBdyWidget(self, scope=scope, name=self.name)

materialparameter.ListOfInterfacesSkelBdyParameter.makeWidget = _makeListOfInterfacesSkelBdyWidget

#Listbox. This widget differs from the above widget by having
#a self.interfacewidget member
class ListOfInterfacesSkelBdyWithMaterialWidget(ListOfInterfacesWithMaterialWidget):
    def update(self, *args, **kwargs):
        msname = self.mswidget.get_value()
        interfacematname = self.interfacewidget.get_value()
        if msname:
            ms = ooflib.common.microstructure.microStructures[msname]
            if ms:
                interfacemsplugin=ms.getObject().getPlugIn("Interfaces")
                inames=interfacemsplugin.getInterfaceNamesWithMaterial(interfacematname)
                sbnames=interfacemsplugin.getSkelBdyNamesWithMaterial(interfacematname)
                inames.sort()
                sbnames.sort()
                names=inames+sbnames
                self.widget.update(names)
                self.widgetChanged(len(names) > 0, interactive=0)
                return
        self.widget.update([])
        self.widgetChanged(0, interactive=0)

def _makeListOfInterfacesSkelBdyWithMaterialWidget(self, scope=None):
    return ListOfInterfacesSkelBdyWithMaterialWidget(self, scope=scope, name=self.name)

materialparameter.ListOfInterfacesSkelBdyWithMaterialParameter.makeWidget = _makeListOfInterfacesSkelBdyWithMaterialWidget
