# -*- python -*-
# $RCSfile: interfaceparameters.py,v $
# $Revision: 1.3 $
# $Author: rlua $
# $Date: 2007-10-01 19:24:27 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from ooflib.common import enum
from ooflib.common import utils

#Interface branch

class MaterialType(enum.EnumClass(
    ('bulk', 'A bulk material. Normal bulk properties can be added only to this type of material.'),
    ('interface', 'An interface material. Interface properties can be added only to this type of material.')
    )):
    tip = "Types of &oof2; materials."
##    discussion = "<para>See <xref linkend='Section:Concepts:FileFormats'/>.</para>"

utils.OOFdefine('MaterialType', MaterialType)

MATERIALTYPE_BULK = MaterialType("bulk")
MATERIALTYPE_INTERFACE = MaterialType("interface")
#Might be able to combine the material type and compatibilty
#if we can imagine a material that accepts both bulk and interface
#types.

#These are used in PropertyRegistration. Could turn these
#into enums later.
##COMPATIBILITY_BULK_ONLY = 0
##COMPATIBILITY_INTERFACE_ONLY = 1
##COMPATIBILITY_BULK_INTERFACE = 2

class MatPropCompatibilityType(enum.EnumClass(
    ('bulk_only', 'The property can only be assigned to bulk materials.'),
    ('interface_only', 'The property can only be assigned to interface materials.'),
    ('bulk_interface', 'The property can be assigned to either bulk or interface materials.')
    )):
    tip = "Compatibility of &oof2; properties to &oof2; materials."
##    discussion = "<para>See <xref linkend='Section:Concepts:FileFormats'/>.</para>"

utils.OOFdefine('MatPropCompatibilityType', MatPropCompatibilityType)

COMPATIBILITY_BULK_ONLY = MatPropCompatibilityType('bulk_only')
COMPATIBILITY_INTERFACE_ONLY = MatPropCompatibilityType('interface_only')
COMPATIBILITY_BULK_INTERFACE = MatPropCompatibilityType('bulk_interface')
