# -*- python -*-
# $RCSfile: interfaceplugin.py,v $
# $Revision: 1.9 $
# $Author: rlua $
# $Date: 2007-10-25 19:15:14 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from ooflib.SWIG.common import switchboard
from ooflib.SWIG.common import pixelgroup
from ooflib.SWIG.common import ooferror
from ooflib.common import microstructure
from ooflib.common import labeltree
from ooflib.common import registeredclass
from ooflib.common import utils
from ooflib.common.IO import pixelgroupparam
from ooflib.common.IO import microstructureIO
from ooflib.common.IO import mainmenu
from ooflib.engine import skeletonelement
from ooflib.engine import subproblemcontext
from ooflib.engine import bdycondition
from ooflib.engine import materialmanager
from ooflib.engine import skeletoncontext
from ooflib.engine.IO import interfaceparameters
from ooflib.engine.IO import materialparameter
from ooflib.engine.IO import materialmenu
import ooflib.engine.mesh

OOF = mainmenu.OOF

class InterfaceMSPlugIn(microstructure.MicrostructurePlugIn):
    def __init__(self, ms):
        microstructure.MicrostructurePlugIn.__init__(self, ms)
##        self._namedinterfaces={'interfacename1':MaterialInterface,
##                               'interfacename2':MaterialInterface,
##                               'interfacename3':MaterialInterface,
##                               'interfacename4':MaterialInterface,
##                               'interfacename5':MaterialInterface
##                               }
##        self._materialassignments={'interfacemat1':['interfacename1','interfacename2'],
##                                   'interfacemat2':['interfacename3'],
##                                   'interfacemat3':[]}
        self._namedinterfaces=utils.OrderedDict()
        self._materialassignments={}
        self._selectedInterfaceName=None

        
        self.sbcallbacks = [
            switchboard.requestCallback("materials changed in microstructure",
                                        self.matChanged),
            switchboard.requestCallback('destroy pixel group',
                                        self.destpixgrp),
            switchboard.requestCallback('changed pixel group', self.changedpixgrp)
            #TODO: Figure this out?
            #switchboard.requestCallback('changed pixel groups', self.chngdgrps)
            ]

        #TODO: Somehow use a timestamp in the mesh for interfaces (see mesh.py)

    #This is a response to an (un)assignment of a material to a microstructure
    def matChanged(self, msobj):
        # switchboard "materials changed ..."
        #TODO: Make this more specific (like check if the material that was assigned
        #or unassigned is part of an interface definition...)
        #Would this clash with the other "materials changed..." handlers?
        if msobj==self.microstructure:
            if len(self._namedinterfaces)>0:
                #A rebuild is not trigggered by changes in the skeleton boundaries
                self.rebuildMeshes()

    def destpixgrp(self, group, ms_name):
        # switchboard 'destroy pixel group'
        self.deleteGroup(ms_name,group)

    #Reaction to "changed pixel group" similar to
    #"materials changed in microstructure"
    def changedpixgrp(self, group, ms_name):
        #Check if group is in the definition of one or more interfaces?
        if ms_name==self.microstructure.name():
            self.matChanged(self.microstructure)

    def deleteGroup(self,ms_name,group):
        if ms_name!=self.microstructure.name():
            return
        if len(self._namedinterfaces)==0:
            return
        interfacenames=self.getInterfaceNames()
        doomedinterfaces=[]
        for interfacename in interfacenames:
            interfaceobj=self._namedinterfaces[interfacename]
            if interfaceobj.hasGroup(group.name()):
                del self._namedinterfaces[interfacename]
                doomedinterfaces.append(interfacename)
        self.removeMaterialFromInterfaces(doomedinterfaces)
        #Remove boundary conditions that refer to the doomedinterfaces
        meshclass=ooflib.engine.mesh.meshes
        msname=self.microstructure.name()
        for meshkey in meshclass.keys(base=msname):
            meshctxt=meshclass[[msname]+meshkey]
            for doomedinterfacename in doomedinterfaces:
                meshctxt.removeInterface(doomedinterfacename)
        self.rebuildMeshes()
        switchboard.notify("interface removed",self.microstructure)

    def renameGroup(self,oldgroupname,newgroupname):
        for interfaceobj in self._namedinterfaces.values():
            interfaceobj.renameGroup(oldgroupname,newgroupname)
        #A group has been renamed, it is possible the details of
        #one or more interfaces have been changed. 
        switchboard.notify("interface renamed",self.microstructure)

    def destroy(self):
        map(switchboard.removeCallback, self.sbcallbacks)
        microstructure.MicrostructurePlugIn.destroy(self)

    def interfaceInfo(self,interfacename):
        try:
            return self._namedinterfaces[interfacename].__repr__()
        except KeyError:
            return ""

    def getInterfaceNames(self):
        return self._namedinterfaces.keys()[:]

    #Returns strings of the form "skeleton:bdkey"
    def getSkelBdyNames(self):
        msname=self.microstructure.name()
        skelclass=skeletoncontext.skeletonContexts
        names=[]
        for skelkey in skelclass.keys(base=msname):
            skelctxt=skelclass[[msname]+skelkey]
            for bname in skelctxt.allEdgeBoundaryNames():
                names.append(skelkey[0]+":"+bname)
        return names

    def getInterfaceNamesWithMaterial(self,matname):
        try:
            names=self._materialassignments[matname][:]
            return names
        except KeyError:
            return []

    def getSkelBdyNamesWithMaterial(self,matname):
        msname=self.microstructure.name()
        skelclass=skeletoncontext.skeletonContexts
        names=[]
        for skelkey in skelclass.keys(base=msname):
            skelctxt=skelclass[[msname]+skelkey]
            for bname in skelctxt.allEdgeBoundaryNames():
                bdyctxt=skelctxt.getBoundary(bname)
                if bdyctxt._interfacematerial==matname:
                    names.append(skelkey[0]+":"+bname)
        return names

    #Get material assigned to interfacename
    def getInterfaceMaterialName(self,interfacename):
        for matname, interfacenamelist in self._materialassignments.items():
            if interfacename in interfacenamelist:
                return matname
        return None

    #This method returns materials assigned to interfaces and boundaries.
    #In contrast, materialmanager.getInterfaceMaterialNames returns
    #all interface materials listed on the Materials page.
    def getInterfaceMaterials(self):
        matdict={}
        for matname, interfacenames in self._materialassignments.items():
            if len(interfacenames)>0:
                matdict[matname]=1
        msname=self.microstructure.name()
        skelclass=skeletoncontext.skeletonContexts
        for skelkey in skelclass.keys(base=msname):
            skelctxt=skelclass[[msname]+skelkey]
            for bname in skelctxt.allEdgeBoundaryNames():
                bdyctxt=skelctxt.getBoundary(bname)
                if bdyctxt._interfacematerial!=None:
                    matdict[bdyctxt._interfacematerial]=1
        return matdict.keys()

    def addInterface(self,name,interfacedef):
        self._namedinterfaces[name]=interfacedef
        switchboard.notify("new interface created",self.microstructure)
        self.selectInterface(name)
        self.rebuildMeshes()

    def removeInterface(self,name):
        try:
            del self._namedinterfaces[name]
            interfacenames=self.getInterfaceNames()
            doomedinterfaces=[name]
            #Because of the possibility of compound interfaces (e.g. union),
            #we must search the other interfaces for the presence
            #of the interface corresponding to name.
            for interfacename in interfacenames:
                interfaceobj=self._namedinterfaces[interfacename]
                if interfaceobj.hasInterface(name):
                    del self._namedinterfaces[interfacename]
                    doomedinterfaces.append(interfacename)
            self.removeMaterialFromInterfaces(doomedinterfaces)
            self.unselectInterface()
            #Remove boundary conditions that refer to the doomedinterfaces
            meshclass=ooflib.engine.mesh.meshes
            msname=self.microstructure.name()
            for meshkey in meshclass.keys(base=msname):
                meshctxt=meshclass[[msname]+meshkey]
                for doomedinterfacename in doomedinterfaces:
                    meshctxt.removeInterface(doomedinterfacename)
            self.rebuildMeshes()
            switchboard.notify("interface removed",self.microstructure)
        except KeyError:
            pass

    def getCurrentReservedNames(self):
        reservednamedict={}
        for iname in self.getInterfaceNames():
            reservednamedict[iname]=1
        skelclass=skeletoncontext.skeletonContexts
        msname=self.microstructure.name()
        for skelkey in skelclass.keys(base=msname):
            skelctxt=skelclass[[msname]+skelkey]
            for bname in skelctxt.allBoundaryNames():
                reservednamedict[bname]=1
        reservednamedict['top']=1
        reservednamedict['bottom']=1
        reservednamedict['left']=1
        reservednamedict['right']=1
        reservednamedict['topleft']=1
        reservednamedict['topright']=1
        reservednamedict['bottomleft']=1
        reservednamedict['bottomright']=1
        return reservednamedict.keys()

    def renameInterface(self,oldname,newname):
        if oldname==newname:
            return
        if newname in self.getCurrentReservedNames():
            raise ooferror.ErrSetupError("Name %s already in use." % newname)
        try:
            obj=self._namedinterfaces[oldname]
            del self._namedinterfaces[oldname]
            self._namedinterfaces[newname]=obj
            interfacenames=self.getInterfaceNames()
            #Because of the possibility of compound interfaces (e.g. union),
            #we must search the other interfaces for the presence
            #of the interface corresponding to oldname.
            for interfacename in interfacenames:
                interfaceobj=self._namedinterfaces[interfacename]
                interfaceobj.renameInterface(oldname,newname)
            #interface must also be renamed in the list of
            #material assignments
            matname=self.getInterfaceMaterialName(oldname)
            self.removeMaterialFromInterfaces([oldname])
            self.assignMaterialToInterfaces(matname,[newname])
            #Rename the interface in the boundary conditions
            #and in the edgements.
            meshclass=ooflib.engine.mesh.meshes
            msname=self.microstructure.name()
            for meshkey in meshclass.keys(base=msname):
                meshctxt=meshclass[[msname]+meshkey]
                meshctxt.renameInterface(oldname,newname)
            self.selectInterface(newname)
            switchboard.notify("interface renamed",self.microstructure)
        except KeyError:
            pass

    #Assign material (matname) to interfaces and boundaries in the list interfacenames
    def assignMaterialToInterfaces(self,matname,interfacenames):
        
        #Extract the skeleton boundary names from interfacenames.
        #The skeleton boundaries are not listed in _materialassignments.
        #The interfacematerial is an attribute of the
        #skeleton context edge boundary. This way,
        #we do not have to explicitly handle changes in the
        #names of skeletons and skeleton boundaries.
        msname=self.microstructure.name()
        skelclass=skeletoncontext.skeletonContexts
        for skelkey in skelclass.keys(base=msname):
            skelctxt=skelclass[[msname]+skelkey]
            for bname in skelctxt.allEdgeBoundaryNames():
                longname=skelkey[0]+":"+bname
                if longname in interfacenames:
                    interfacenames.remove(longname)
                    bdyctxt=skelctxt.getBoundary(bname)
                    bdyctxt._interfacematerial=matname

        #First unassign interfacenames from any material
        self.removeMaterialFromInterfaces(interfacenames)
        try:
            interfacenamelist=self._materialassignments[matname]
            #After the call to removeMaterialFromInterfaces above,
            #there should be no duplicates in the sum on the right hand side
            self._materialassignments[matname]=interfacenamelist+interfacenames
        except KeyError:
            self._materialassignments[matname]=interfacenames[:]

    #Remove the interfaces indicated in interfacenames from their material
    #assignments
    def removeMaterialFromInterfaces(self,interfacenames):
        #The skeleton boundaries are not listed in _materialassignments.
        #The interfacematerial is an attribute of the
        #skeleton context edge boundary. This way,
        #we do not have to explicitly handle changes in the
        #names of skeletons and skeleton boundaries.
        msname=self.microstructure.name()
        skelclass=skeletoncontext.skeletonContexts
        for skelkey in skelclass.keys(base=msname):
            skelctxt=skelclass[[msname]+skelkey]
            for bname in skelctxt.allEdgeBoundaryNames():
                longname=skelkey[0]+":"+bname
                if longname in interfacenames:
                    #interfacenames.remove(longname)
                    bdyctxt=skelctxt.getBoundary(bname)
                    bdyctxt._interfacematerial=None

        for matname, interfacenamelist in self._materialassignments.items():
            for interfacename in interfacenames:
                if interfacename in interfacenamelist:
                    interfacenamelist.remove(interfacename)

    #Implement changes to the interfaces when a material is deleted
    #in the materials page. If the material is an interface material,
    #remove entry for this material in _materialassignments.
    #If the material is a bulk material, interfaces that have this
    #material as an attribute are removed.
    #TODO: For a union of interfaces, not necessary to delete the whole
    #union.
    def deleteMaterial(self,matname,material_type):
        if material_type==interfaceparameters.MATERIALTYPE_INTERFACE.name:
            try:
                del self._materialassignments[matname]
            except KeyError:
                pass
            #Unassociate matname from skeleton boundaries
            msname=self.microstructure.name()
            skelclass=skeletoncontext.skeletonContexts
            for skelkey in skelclass.keys(base=msname):
                skelctxt=skelclass[[msname]+skelkey]
                for bname in skelctxt.allEdgeBoundaryNames():
                    bdyctxt=skelctxt.getBoundary(bname)
                    if bdyctxt._interfacematerial==matname:
                        bdyctxt._interfacematerial=None
        else:
            if len(self._namedinterfaces)==0:
                return
            interfacenames=self.getInterfaceNames()
            doomedinterfaces=[]
            for interfacename in interfacenames:
                interfaceobj=self._namedinterfaces[interfacename]
                if interfaceobj.hasBulkMaterial(matname):
                    del self._namedinterfaces[interfacename]
                    doomedinterfaces.append(interfacename)
            self.removeMaterialFromInterfaces(doomedinterfaces)
            #Remove boundary conditions that refer to the doomedinterfaces
            meshclass=ooflib.engine.mesh.meshes
            msname=self.microstructure.name()
            for meshkey in meshclass.keys(base=msname):
                meshctxt=meshclass[[msname]+meshkey]
                for doomedinterfacename in doomedinterfaces:
                    meshctxt.removeInterface(doomedinterfacename)
            self.rebuildMeshes()
            switchboard.notify("interface removed",self.microstructure)

    def renameMaterial(self,oldmatname,newmatname,material_type):
        if material_type==interfaceparameters.MATERIALTYPE_INTERFACE.name:
            try:
                interfacenamelist=self._materialassignments[oldmatname]
                del self._materialassignments[oldmatname]
                self._materialassignments[newmatname]=interfacenamelist
            except KeyError:
                pass
            #Rename materials associated with skeleton boundaries
            msname=self.microstructure.name()
            skelclass=skeletoncontext.skeletonContexts
            for skelkey in skelclass.keys(base=msname):
                skelctxt=skelclass[[msname]+skelkey]
                for bname in skelctxt.allEdgeBoundaryNames():
                    bdyctxt=skelctxt.getBoundary(bname)
                    if bdyctxt._interfacematerial==oldmatname:
                        bdyctxt._interfacematerial=newmatname
        else:
            for interfaceobj in self._namedinterfaces.values():
                interfaceobj.renameBulkMaterial(oldmatname,newmatname)
            #A material has been renamed, it is possible the details of
            #one or more interfaces have been changed. 
            switchboard.notify("interface renamed",self.microstructure)

    # The selected interface isn't actually used for anything directly,
    # but it's displayed in the GUI and its name is passed in as an
    # argument to commands that operate on interfaces.  
    def selectInterface(self, interfacename):
        self._selectedInterfaceName = interfacename
        #self.bdyselected.increment()    # timestamp
        switchboard.notify("interface selected",
                           self.microstructure, #ms object (microstructure.py)
                           interfacename)
        #switchboard.notify('redraw')
    def unselectInterface(self):
        if self._selectedInterfaceName is not None:
            self._selectedInterfaceName = None
            #self.bdyselected.increment()
            switchboard.notify("interface unselected",
                               self.microstructure)
            #switchboard.notify('redraw')
    def getSelectedInterfaceName(self):
        return self._selectedInterfaceName
    def getSelectedInterface(self):      # returns Interface object
        if self._selectedInterfaceName is not None:
            return self._namedinterfaces[self._selectedInterfaceName]

    #Cases when we want to rebuild the mesh:
    #(1) Interfaces added or removed
    #(2) Bulk materials get deleted
    #(3) Bulk materials get assigned to pixels (not handled yet)
    #(4) Renaming an interface material might also trigger a rebuild(?)
    def rebuildMeshes(self):
        print "rebuildMeshes"
        #Alternative to ooflib.engine.mesh.meshes: whoville.getClass("Mesh")
        meshclass=ooflib.engine.mesh.meshes
        msname=self.microstructure.name()
        for meshkey in meshclass.keys(base=msname):
            self.rebuildMesh(meshclass[[msname]+meshkey])
    def rebuildMesh(self,meshcontext):
        #Most of the following snippet taken from adaptivemeshrefinement.py
        meshcontext.reserve() #Necessary?
        meshcontext.begin_writing()

        old_femesh = meshcontext.femesh()
        old_skel = meshcontext.getSkeleton()
        # realmaterial is a function which tells the elements of the
        # created femesh how to get their materials.
        realmaterial = skeletonelement.SkeletonElement.realmaterial
        #Interface branch, pass skeleton path to femesh
        skelpath=labeltree.makePath(meshcontext.path())
        new_femesh = old_skel.femesh(meshcontext.elementdict,
                                     realmaterial,
                                     skelpath[0]+":"+skelpath[1])
        new_femesh.set_rwlock(meshcontext.rwLock)

        
        # This code is vaguely similar to that in copyMesh in
        # meshmenu.py (the callback for OOF.Mesh.Copy), but is *not*
        # identical.  That's because in this case we're replacing the
        # FEMesh inside a Mesh context, instead of simply creating a
        # whole new Mesh context.

        # Copy subproblems, their fields and equations.  The tricky
        # bit here is that the new subproblems must live in the same
        # Mesh as the old ones and must have the same names, but they
        # refer to different FEMeshes.  If the old subproblems are
        # deleted before the new ones are created, then their Field
        # values in the old FEMesh will be deleted before they can be
        # copied to the new FEMesh.  On the other hand, if the old
        # subproblems are kept, then their names will conflict with
        # the new ones.  Therefore, we have to create the new
        # subproblems with altered names, and change the names after
        # the old subproblems are deleted.  Also, we can't actually
        # insert the new subproblems into the Mesh context until the
        # new FEMesh has been installed...

        # First, gather data about subproblems and their fields and
        # equations.
        class SubProbState:
            pass
        subpstates = {}
        for old_subpctxt in meshcontext.subproblems():
            old_subp = old_subpctxt.getObject()
            subpstate = subpstates[old_subpctxt.path()] = SubProbState()
            subpstate.old_subp = old_subp
            subpstate.new_subp = old_subp.clone()
            subpstate.new_subp.set_time(old_subp.current_time())
            subpstate.fields = old_subp.all_fields() # all defined fields
            subpstate.activefields = [f for f in subpstate.fields
                                      if old_subp.is_active_field(f)]
            subpstate.eqns = old_subp.all_equations() # all active equations
            
        # replace old_femesh
        meshcontext.setFemesh(new_femesh)

        # Install new subproblems and their fields and equations
        for subppath, subpstate in subpstates.items():
            tempname = subproblemcontext.subproblems.uniqueName(subppath)
            subpstate.tempname = tempname
            #Enclosing between pause_writing and resume_writing fixes
            #a blocking problem. May not be the final word...
            meshcontext.pause_writing()
            meshcontext.newSubProblem(subpstate.new_subp,
                                      meshcontext.path() + ":" + tempname)
            meshcontext.resume_writing()
            for field in subpstate.fields:
                subpstate.new_subp.define_field(field)
                new_femesh.set_in_plane(field, old_femesh.in_plane(field))
            for field in subpstate.activefields:
                subpstate.new_subp.activate_field(field)
            for equation in subpstate.eqns:
                subpstate.new_subp.activate_equation(equation)

        # Copy Field values from the old FEMesh
        for field in meshcontext.all_subproblem_fields():
            new_femesh.init_field(old_skel, old_femesh, field)

        # Destroy old subproblems (this has to happen *after* fields
        # are copied) and give new subproblems their correct names.
        meshcontext.pause_writing()
        try:
            for subppath, subpstate in subpstates.items():
                oldname = labeltree.makePath(subppath)[-1]
                subpctxt = subproblemcontext.subproblems[subppath]
                subpctxt.begin_writing()
                try:
                    subpctxt.destroy()
                finally:
                    subpctxt.end_writing()
                temppath = meshcontext.path() + ':' + subpstate.tempname
                newctxt = subproblemcontext.subproblems[temppath]
                newctxt.begin_writing()
                try:
                    newctxt.rename(oldname)
                finally:
                    newctxt.end_writing()
                ## Passing the default subproblem to the FEMesh is a hack,
                ## intended to be removed once adaptive mesh refinement is
                ## moved from FEMesh to CSubProblem.
                #if oldname == 'default':
                #    new_femesh.setDefaultSubProblem(newctxt.getObject())
                    
        finally:
            meshcontext.resume_writing()

        # put BC in new_femesh
        for (name, bc) in meshcontext.allBoundaryConds():
            if name.find('_cntnty_')==0:
                #These mesh boundaries are not derived from skeleton boundaries
                #(see femesh.spy). They are mesh boundaries created for the
                #purpose of setting Float BCs that enforce (dis)continuity of
                #fields across an interface.
                continue
            if isinstance(bc,bdycondition.JumpBC):
                #For JumpBC, bc.boundary is just the name of the interface.
                #There is no mesh boundary associated with it.
                #The actual boundary condition object is still
                #with meshcontext
                new_femesh.addInterfaceBC(name)
                continue
            bc.boundary_obj=new_femesh.getBoundary(bc.boundary)
            new_femesh.boundaries[bc.boundary].addCondition(bc)
        meshcontext.pause_writing()
        try:
            for subproblem in meshcontext.subproblems():
                subproblem.autoenableBCs()
        finally:
            meshcontext.resume_writing()
            
        old_femesh.destroy()
                
        # time stamp update
        meshcontext.changed()

        meshcontext.end_writing()
        meshcontext.cancel_reservation()

        switchboard.notify("mesh changed", meshcontext)
        switchboard.notify('redraw')
        #switchboard.notify('Mesh modified', mesh, modifier)
        switchboard.notify('mesh boundaries changed', meshcontext.getObject())

microstructure.registerMicrostructurePlugIn(InterfaceMSPlugIn, "Interfaces")

#######################################################
#Interface definitions

class InterfaceDef(registeredclass.RegisteredClass):
    registry = []
    def addToMS(self,interfacemsplugin,interfacename):
        interfacemsplugin.addInterface(interfacename,self)
    def check(self):
        pass
    def hasBulkMaterial(self,bulkmatname):
        return 0
    def renameBulkMaterial(self,oldname,newname):
        return 0
    #hasGroup and renameGroup are isomorphic to
    #hasBulkMaterial and renameBulkMaterial.
    #A generic function may be defined that does
    #hasGroup and hasBulkMaterial, for example,
    #but I'll leave the distinction here.
    def hasGroup(self,groupname):
        return 0
    def renameGroup(self,oldname,newname):
        return 0
    def hasInterface(self,interfacename):
##        if interfaceobj==self:
##            return 1
        return 0
    def renameInterface(self,oldname,newname):
        pass

class MaterialInterface(InterfaceDef):
    def __init__(self,left,right):
        #members must be spelled the same as the parameters
        self.left=left
        self.right=right
    def check(self):
        if self.left==self.right:
            return 'The two sides must have different materials!'
    #RegisteredClass already has a __repr__
##    def info(self):
##        return "Between Bulk Materials (%s,%s)" % (self.left,self.right)
    #Useful function for finding out if the pair of bulk materials
    #already has an interface material assigned to it.
##    def __eq__(self,other):
##        if self.__class__!=other.__class__:
##            return 0
##        return ((self._mat1==other._mat1 and \
##                 self._mat2==other._mat2) or
##                (self._mat1==other._mat2 and
##                 self._mat2==other._mat1))
    #This function is used by the mesh construction function to test
    #whether a skeleton segment is part of the interface.
    def isInterfaceSegment(self,seg,skelobj):
        els = seg.getElements()
        if len(els) > 1:
            mat1 = els[0].material(skelobj)
            mat2 = els[1].material(skelobj)
            #No need to perform this check
##            if mat1==mat2:
##                return (0,None)
            mat1name="<No material>"
            if mat1:
                mat1name=mat1.name()
            mat2name="<No material>"
            if mat2:
                mat2name=mat2.name()
            if (self.left==mat1name and self.right==mat2name):
                return (1,els[0])
            if (self.right==mat1name and self.left==mat2name):
                return (1,els[1])
        else:
            mat = els[0].material(skelobj)
            #Treat region outside of mesh as having "<No material>"
            if mat:
                if (self.left==mat.name() and self.right=="<No material>"):
                    return (1,els[0])
                if (self.right==mat.name() and self.left=="<No material>"):
                    return (1,els[0]) #I could pass "(1,None)" here.
        return (0,None)
    def hasBulkMaterial(self,bulkmatname):
        if self.left==bulkmatname or self.right==bulkmatname:
            return 1
        return 0
    def renameBulkMaterial(self,oldname,newname):
        numrenamed=0
        if self.left==oldname:
            self.left=newname
            numrenamed+=1
        if self.right==oldname:
            self.right=newname
            numrenamed+=1
        return 0

registeredclass.Registration(
    "Between Bulk Materials",
    InterfaceDef,
    MaterialInterface,
    ordering=10,
    params = [
    materialparameter.BulkMaterialParameterExtra('left',
                                                 tip="Bulk Material."),
    materialparameter.BulkMaterialParameterExtra('right',
                                                 tip="Bulk Material.")
    ]
    )

class SingleMaterialInterface(InterfaceDef):
    def __init__(self,left):
        self.left=left
    #This function is used by the mesh construction function to test
    #whether a skeleton segment is part of the interface.
    def isInterfaceSegment(self,seg,skelobj):
        els = seg.getElements()
        if len(els) > 1:
            mat1 = els[0].material(skelobj)
            mat2 = els[1].material(skelobj)
            if mat1==mat2:
                return (0,None)
            if self.left=="<No material>":
                if (mat1 is None):
                    return (1,els[0])
                if (mat2 is None):
                    return (1,els[1])
            if (mat1 and self.left==mat1.name()):
                return (1,els[0])
            if (mat2 and self.left==mat2.name()):
                return (1,els[1])
        else:
            mat = els[0].material(skelobj)
            if mat:
                if self.left==mat.name():
                    return (1,els[0])
        return (0,None)
    def hasBulkMaterial(self,bulkmatname):
        if self.left==bulkmatname:
            return 1
        return 0
    def renameBulkMaterial(self,oldname,newname):
        if self.left==oldname:
            self.left=newname
            return 1
        return 0

registeredclass.Registration(
    "Around a Single Material",
    InterfaceDef,
    SingleMaterialInterface,
    ordering=15,
    params = [
    materialparameter.BulkMaterialParameterExtra('left',
                                                 tip="Bulk Material.")
    ]
    )

class PixelGroupInterface(InterfaceDef):
    def __init__(self,left,right):
        self.left=left
        self.right=right
    def check(self):
        if self.left==self.right:
            return 'The two sides must have different pixel groups!'
    #This function is used by the mesh construction function to test
    #whether a skeleton segment is part of the interface.
    def isInterfaceSegment(self,seg,skelobj):
        els = seg.getElements()
        if len(els) > 1:
            msobj=skelobj.MS
            where = msobj.getRepresentativePixel(
                els[0].dominantPixel(msobj))
            grpnames1 = pixelgroup.pixelGroupNames(msobj, where)
            where = msobj.getRepresentativePixel(
                els[1].dominantPixel(msobj))
            grpnames2 = pixelgroup.pixelGroupNames(msobj, where)
            if self.left!="<No pixelgroup>" and \
                   self.right!="<No pixelgroup>":
                if (self.left in grpnames1 and \
                    self.right in grpnames2):
                    return (1,els[0])
                if (self.left in grpnames2 and \
                    self.right in grpnames1):
                    return (1,els[1])
            elif self.left!="<No pixelgroup>":
                if (self.left in grpnames1 and \
                    len(grpnames2)==0):
                    return (1,els[0])
                if (self.left in grpnames2 and \
                    len(grpnames1)==0):
                    return (1,els[1])
            elif self.right!="<No pixelgroup>":
                if (self.right in grpnames1 and \
                    len(grpnames2)==0):
                    return (1,els[1])
                if (self.right in grpnames2 and \
                    len(grpnames1)==0):
                    return (1,els[0])
        else:
            msobj=skelobj.MS
            where = msobj.getRepresentativePixel(
                els[0].dominantPixel(msobj))
            grpnames = pixelgroup.pixelGroupNames(msobj, where)
            #Treat region outside of mesh as having "<No pixelgroup>"
            if self.left=="<No pixelgroup>" and \
                   (self.right in grpnames):
                return (1,els[0])
            if self.right=="<No pixelgroup>" and \
                   (self.left in grpnames):
                return (1,els[0])
        return (0,None)
    def hasGroup(self,groupname):
        if self.left==groupname or self.right==groupname:
            return 1
        return 0
    def renameGroup(self,oldname,newname):
        if self.left==oldname:
            self.left=newname
            return 1
        if self.right==oldname:
            self.right=newname
            return 1
        return 0

registeredclass.Registration(
    "Between Pixel Groups",
    InterfaceDef,
    PixelGroupInterface,
    ordering=20,
    params = [
    pixelgroupparam.PixelGroupInterfaceParameter('left',
                                                 tip="Pixels."),
    pixelgroupparam.PixelGroupInterfaceParameter('right',
                                                 tip="Pixels.")
    ]
    )

class SinglePixelGroupInterface(InterfaceDef):
    def __init__(self,left):
        self.left=left
    #This function is used by the mesh construction function to test
    #whether a skeleton segment is part of the interface.
    def isInterfaceSegment(self,seg,skelobj):
        els = seg.getElements()
        if len(els) > 1:
            msobj=skelobj.MS
            where = msobj.getRepresentativePixel(
                els[0].dominantPixel(msobj))
            grpnames1 = pixelgroup.pixelGroupNames(msobj, where)
            where = msobj.getRepresentativePixel(
                els[1].dominantPixel(msobj))
            grpnames2 = pixelgroup.pixelGroupNames(msobj, where)
            if self.left!="<No pixelgroup>":
                if (self.left in grpnames1 and \
                    self.left not in grpnames2):
                    return (1,els[0])
                if (self.left in grpnames2 and \
                    self.left not in grpnames1):
                    return (1,els[1])
            else:
                if (len(grpnames1)>0 and len(grpnames2)==0):
                    return (1,els[1])
                if (len(grpnames2)>0 and len(grpnames1)==0):
                    return (1,els[0])
        else:
            msobj=skelobj.MS
            where = msobj.getRepresentativePixel(
                els[0].dominantPixel(msobj))
            grpnames = pixelgroup.pixelGroupNames(msobj, where)
            #if self.left!="<No pixelgroup>":
            if self.left in grpnames:
                return (1,els[0])
        return (0,None)
    def hasGroup(self,groupname):
        if self.left==groupname:
            return 1
        return 0
    def renameGroup(self,oldname,newname):
        if self.left==oldname:
            self.left=newname
            return 1
        return 0

registeredclass.Registration(
    "Around a Pixel Group",
    InterfaceDef,
    SinglePixelGroupInterface,
    ordering=30,
    params = [
    pixelgroupparam.PixelGroupInterfaceParameter('left',
                                                 tip="Pixels.")
    ]
    )

#Create a CompoundInterface parent class?
class UnionInterface(InterfaceDef):
    def __init__(self,interfaces):
        self.interfaces=interfaces
        self._namedinterfaces=utils.OrderedDict()
##        for interfacename in interfaces:
##            self._namedinterfaces[interfacename]=interfacemsplugin._namedinterfaces[interfacename]
    def addToMS(self,interfacemsplugin,interfacename):
        for name in self.interfaces:
            self._namedinterfaces[name]=interfacemsplugin._namedinterfaces[name]
        interfacemsplugin.addInterface(interfacename,self)
##    def info(self):
##        return "Union of interfaces\n" + self._namedinterfaces
##    def __eq__(self,other):
##        if self.__class__!=other.__class__:
##            return 0
##        return ((self._interface1==other._interface1 and \
##                 self._interface2==other._interface2) or
##                (self._interface1==other._interface2 and
##                 self._interface2==other._interface1))
    #This function is used by the mesh construction function to test
    #whether a skeleton segment is part of the interface.
    def isInterfaceSegment(self,seg,skelobj):
        for interfaceobj in self._namedinterfaces.values():
            (yes, el)=interfaceobj.isInterfaceSegment(seg,skelobj)
            if yes:
                return (1, el)
        return (0,None)
    def hasBulkMaterial(self,bulkmatname):
        for interfaceobj in self._namedinterfaces.values():
            if interfaceobj.hasBulkMaterial(bulkmatname):
                return 1
        return 0
    def renameBulkMaterial(self,oldname,newname):
        numrenamed=0
        for interfaceobj in self._namedinterfaces.values():
            if interfaceobj.renameBulkMaterial(oldname,newname):
                numrenamed+=1
        return numrenamed
    def hasInterface(self,interfacename):
        for name, obj in self._namedinterfaces.items():
            if obj.hasInterface(interfacename):
                return 1
            if name==interfacename:
                return 1
        return 0
    def renameInterface(self,oldname,newname):
        interfacenames=self._namedinterfaces.keys()[:]
        for name in interfacenames:
            obj=self._namedinterfaces[name]
            obj.renameInterface(oldname,newname)
            if name==oldname:
                del self._namedinterfaces[oldname]
                self._namedinterfaces[newname]=obj

registeredclass.Registration(
    "Union of named interfaces",
    InterfaceDef,
    UnionInterface,
    ordering=100,
    params = [
    materialparameter.ListOfInterfacesParameter("interfaces",
                                                tip="Current list of interfaces defined in this microstructure."),
    ]
    )

## Define a Microstructure IO PlugIn so that Interfaces
## will be written to Microstructure data files.

#OOF.LoadData.Microstructure.Interface.New is defined in interfacemenu.py
#OOF.LoadData.Material.Interface.Assign is defined in materialmenu.py

def writeInterface(datafile, mscontext):
    #Interface branch
    #TODO: The bulk materials that define the interfaces must be written,
    #even if the material is not assigned to pixels.
    #Interface materials are not assigned to pixels. They must be
    #explicitly listed here.
    msobj=mscontext.getObject()
    interfacemsplugin=msobj.getPlugIn("Interfaces")
    for interfacename, interfacedef in interfacemsplugin._namedinterfaces.items():
        datafile.startCmd(OOF.LoadData.Microstructure.Interface.New)
        datafile.argument('microstructure',msobj.name())
        datafile.argument('name',interfacename)
        datafile.argument('interface_type',interfacedef)
        datafile.endCmd()

    #Write interface properties and materials.
    #An interface material is saved only if it has been assigned to
    #an interface.
    materialmenu.writeMaterials(datafile,
                                [materialmanager.getMaterial(m)
                                 for m in interfacemsplugin.getInterfaceMaterials()])

    #Assign interface materials to interfaces
    for matname, interfacenames in interfacemsplugin._materialassignments.items():
        if len(interfacenames)==0:
            continue
##        datafile.startCmd(OOF.LoadData.MaterialandType)
##        datafile.argument('name',matname)
##        matobj=materialmanager.getMaterial(matname)
##        datafile.argument('properties', [prop.registration().name()
##                                         for prop in matobj.properties()])
##        datafile.argument('materialtype',matobj.type())
##        datafile.endCmd()
        datafile.startCmd(OOF.LoadData.Material.Interface.Assign)
        datafile.argument('microstructure',msobj.name())
        datafile.argument('material',matname)
        datafile.argument('interfaces',interfacenames)
        datafile.endCmd()

microstructureIO.registerMicrostructureIOPlugIn_last(writeInterface)
