# -*- python -*-
# $RCSfile: revertmesh.py,v $
# $Revision: 1.12 $
# $Author: rlua $
# $Date: 2007-10-29 15:47:17 $


# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


from ooflib.common import registeredclass
from ooflib.common import labeltree
from ooflib.common.IO import parameter
from ooflib.engine import errorestimator
from ooflib.engine import meshmod
from ooflib.engine import skeletonelement
from ooflib.engine import refine
from ooflib.engine import subproblemcontext
import string
#Interface branch
from ooflib.engine import bdycondition

# AdaptiveMeshRefinement
class RevertMesh(meshmod.MeshModification):
    def __init__(self, base_skeleton, init_field=1):
        self.base_skeleton = base_skeleton
        self.init_field = init_field
        
##    def apply(self, meshcontext):
##        old_femesh = meshcontext.femesh()  # the current one
##        old_skel = meshcontext.getSkeleton()
##        # Names of base skeletons are automatically generated and
##        # users can't access it.
##        # If there're 3 base skeletons, they'll be
##        # "Base Skeleton 0", "Base Skeleton 1", "Base Skeleton 2".
##        # The numbers in the name string match their positions in the buffer,
##        # which makes retrieving the object simple.
##        ## TODO: What if there are more than 10 skeletons?  This
##        ## retrieval method is *too* simple. (?)
##        index = int(string.split(self.base_skeleton)[-1])
##        # if 'Base Skeleton 0' => 0
##        new_skel = meshcontext.skeleton_buffer.makeCurrent(index)
##        realmaterial = skeletonelement.SkeletonElement.realmaterial
##        new_femesh = new_skel.femesh(meshcontext.elementdict,
##                                     realmaterial,
##                                     self.base_skeleton) #Interface branch, pass skeleton path to femesh
##        new_femesh.set_rwlock(meshcontext.rwLock)
        
##        meshcontext.setSkeleton(new_skel)  # replace old_skel

##        # copy field status.  Only compound fields can be activated/defined.
##        for field in meshcontext.all_compound_subproblem_fields():
##            new_femesh.define_field(field)
##            new_femesh.activate_field(field)
##            if old_femesh.in_plane(field):
##                field.set_in_plane(new_femesh, 1)
        
##        # copy equation status
##        for equation in meshcontext.all_equations():
##            new_femesh.activate_equation(equation)

##        # replace old_femesh
##        meshcontext.setFemesh(new_femesh)

##        # putting BC in new_Femesh
##        for (name, bc) in meshcontext.allBoundaryConds():
##            bc.boundary_obj=new_femesh.getBoundary(bc.boundary)
##            new_femesh.boundaries[bc.boundary].addCondition(bc)
            
##        if self.init_field:
##            # Fields to be initialized from old to new.
##            for field in meshcontext.all_fields():
##                new_femesh.init_field(old_skel, old_femesh, field)
##            ## transfer time index
##            new_femesh.set_time(old_femesh.get_subproblem().current_time())
##        # destroying old_femesh
##        old_femesh.destroy()

##        # time stamp update
##        meshcontext.changed()

    def apply(self, meshcontext):
        old_femesh = meshcontext.femesh()  # the current one
        old_skel = meshcontext.getSkeleton()
        # Names of base skeletons are automatically generated and
        # users can't access it.
        # If there're 3 base skeletons, they'll be
        # "Base Skeleton 0", "Base Skeleton 1", "Base Skeleton 2".
        # The numbers in the name string match their positions in the buffer,
        # which makes retrieving the object simple.
        ## TODO: What if there are more than 10 skeletons?  This
        ## retrieval method is *too* simple. (? I think the split can pick up
        ## more than 1 digit.)
        index = int(string.split(self.base_skeleton)[-1])
        # if 'Base Skeleton 0' => 0
        new_skel = meshcontext.skeleton_buffer.makeCurrent(index)
        realmaterial = skeletonelement.SkeletonElement.realmaterial
        skelpath=labeltree.makePath(meshcontext.path())
        new_femesh = new_skel.femesh(meshcontext.elementdict,
                                     realmaterial,
                                     skelpath[0]+":"+skelpath[1]) #Interface branch, pass skeleton path to femesh
        new_femesh.set_rwlock(meshcontext.rwLock)
        
        meshcontext.setSkeleton(new_skel)  # replace old_skel

        # First, gather data about subproblems and their fields and
        # equations.
        class SubProbState:
            pass
        subpstates = {}
        for old_subpctxt in meshcontext.subproblems():
            old_subp = old_subpctxt.getObject()
            subpstate = subpstates[old_subpctxt.path()] = SubProbState()
            subpstate.old_subp = old_subp
            subpstate.new_subp = old_subp.clone()
            subpstate.new_subp.set_time(old_subp.current_time())
            subpstate.fields = old_subp.all_fields() # all defined fields
            subpstate.activefields = [f for f in subpstate.fields
                                      if old_subp.is_active_field(f)]
            subpstate.eqns = old_subp.all_equations() # all active equations

        # replace old_femesh
        meshcontext.setFemesh(new_femesh)

        # Install new subproblems and their fields and equations
        for subppath, subpstate in subpstates.items():
            tempname = subproblemcontext.subproblems.uniqueName(subppath)
            subpstate.tempname = tempname
            meshcontext.newSubProblem(subpstate.new_subp,
                                      meshcontext.path() + ":" + tempname)
            for field in subpstate.fields:
                subpstate.new_subp.define_field(field)
                new_femesh.set_in_plane(field, old_femesh.in_plane(field))
            for field in subpstate.activefields:
                subpstate.new_subp.activate_field(field)
            for equation in subpstate.eqns:
                subpstate.new_subp.activate_equation(equation)

        # Copy Field values from the old FEMesh
        for field in meshcontext.all_subproblem_fields():
            new_femesh.init_field(old_skel, old_femesh, field)

        # Destroy old subproblems (this has to happen *after* fields
        # are copied) and give new subproblems their correct names.
        meshcontext.pause_writing()
        try:
            for subppath, subpstate in subpstates.items():
                oldname = labeltree.makePath(subppath)[-1]
                subpctxt = subproblemcontext.subproblems[subppath]
                subpctxt.begin_writing()
                try:
                    subpctxt.destroy()
                finally:
                    subpctxt.end_writing()
                temppath = meshcontext.path() + ':' + subpstate.tempname
                newctxt = subproblemcontext.subproblems[temppath]
                newctxt.begin_writing()
                try:
                    newctxt.rename(oldname)
                finally:
                    newctxt.end_writing()
                    
        finally:
            meshcontext.resume_writing()

        # put BC in new_femesh
        for (name, bc) in meshcontext.allBoundaryConds():
            #Interface branch
            #Don't copy the invisible Float BCs associated with
            #interfaces. (see femesh.spy)
            if name.find('_cntnty_')==0:
                continue
            #JumpBC's don't have an associated mesh boundary object
            if isinstance(bc,bdycondition.JumpBC):
                new_femesh.addInterfaceBC(name)
                continue
            bc.boundary_obj=new_femesh.getBoundary(bc.boundary)
            new_femesh.boundaries[bc.boundary].addCondition(bc)

        meshcontext.pause_writing()
        try:
            for subproblem in meshcontext.subproblems():
                subproblem.autoenableBCs()
        finally:
            meshcontext.resume_writing()
            
        old_femesh.destroy()
                
        # time stamp update
        meshcontext.changed()

##        if pBar.query_stop():
##            pBar.set_failure()
##            pBar.set_message("Failed")
##            return None
##        else:
##            pBar.set_progress(1.)
##            pBar.set_message("Done.")

# Trivial parameter for skeletons of the mesh -- it will have
# a custom chooser widget.
class MeshSkeletonParameter(parameter.StringParameter):
    pass

registeredclass.Registration(
    'Revert Mesh',
    meshmod.MeshModification,
    RevertMesh,
    ordering=3,
    params=[
    MeshSkeletonParameter('base_skeleton', tip="Name of the base Skeleton."),
    parameter.BooleanParameter('init_field', 1, default=1,
                               tip="Initialize fields?")
    ],
    tip="Undo adaptive mesh refinement.",
    discussion="""<para>
    Refined &mesh; can be reverted back to its pre-refined stages, in case
    the refinement was not satisfactory. Field values (solution) may not be
    recovered properly in some cases.
    </para>"""
    )
